# Copyright (C) 2024 Canonical Ltd.
# Author: Chris Peterson <chris.peterson@canonical.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

import unittest


# Binary Tests
class BinaryTests(unittest.TestCase):

    # The requestsync binary has the option of using the launchpad api
    # to log in but requires python3-keyring in addition to
    # python3-launchpadlib. Testing the integrated login functionality
    # automatically isn't very feasbile, but we can at least write a smoke
    # test to make sure the required packages are installed.
    # See LP: #2049217
    def test_keyring_installed(self):
        """Smoke test for required lp api dependencies"""
        try:
            # pylint: disable-next=import-outside-toplevel,unused-import
            import keyring  # noqa: F401
        except ModuleNotFoundError as error:
            raise ModuleNotFoundError("package python3-keyring is not installed") from error
